from nltk import word_tokenize, WordNetLemmatizer
from nltk.corpus import stopwords


def preprocess_message(message):
    stop_words = stopwords.words('english')
    tokens = word_tokenize(message)
    tokens = [word.lower() for word in tokens if word not in stop_words and word.isalpha()]
    lemmatizer = WordNetLemmatizer()
    tokens = [lemmatizer.lemmatize(word) for word in tokens]
    return tokens, ' '.join(tokens)



import click

from .rosetta_index import RosettaIndex
from .utils import check_repo_path, normalise_paths, check_index_path, json_echo


class Context:
    def __init__(self):
        self.repo_path = None
        self.index_path = None
        self.json_mode = False


pass_context = click.make_pass_decorator(Context, ensure=True)


@click.group()
@click.option('--json-mode', is_flag=True)
@click.argument('repo-path', type=click.Path())
@click.option('--index-path', type=click.Path())
@pass_context
def rosetta(ctx, json_mode, repo_path, index_path):
    ctx.json_mode = json_mode
    repo_path, index_path = normalise_paths(repo_path, index_path)
    check_repo_path(repo_path)
    index_path = check_index_path(repo_path, index_path)
    ctx.repo_path = repo_path
    ctx.index_path = index_path


@rosetta.command()
@click.option('--update/--noupdate', default=False)
@pass_context
def load(ctx, update):
    index = RosettaIndex(ctx.repo_path, ctx.index_path)
    last_updated = index.get_last_updated()
    if ctx.json_mode:
        json_echo({"status": "success", "last_update_time": f"{last_updated}"})
    else:
        click.echo(f"Index last updated at {last_updated}")


@rosetta.command()
@pass_context
def update(ctx):
    index = RosettaIndex(ctx.repo_path, ctx.index_path)
    last_updated = index.get_last_updated()
    click.echo(f"Most recent update was at {last_updated}")
    num_of_commits, last_updated = index.update_index()
    click.echo(f"{num_of_commits} new commits indexed successfully.")


@rosetta.command()
@click.option('--query', type=click.STRING)
@pass_context
def search(ctx, query):
    index = RosettaIndex(ctx.repo_path, ctx.index_path)
    results = index.search_index(query)
    if ctx.json_mode:
        json_echo({"status": "success", "results": results})
    else:
        click.echo(results)


if __name__ == '__main__':
    rosetta()



import sqlite3
from uuid import uuid4


def uuid():
    return str(uuid4())


class Database:
    def __init__(self, db_path):
        self.db_location = db_path
        self.create_if_not_exists_tables()

    def db_is_empty(self):
        con = sqlite3.connect(self.db_location)
        cur = con.cursor()

        tables = cur.execute("SELECT name FROM sqlite_master WHERE type = 'table' AND name = 'commits'").fetchall()

        con.commit()
        con.close()

        if not tables:
            return True
        else:
            return False

    def index_is_empty(self):
        con = sqlite3.connect(self.db_location)
        cur = con.cursor()

        count = cur.execute("SELECT COUNT(*) FROM commits").fetchone()[0]

        con.commit()
        con.close()

        if count == 0:
            return True
        else:
            return False

    def insert_into_commits(self, commit, message):
        con = sqlite3.connect(self.db_location)
        cur = con.cursor()

        sql = "INSERT INTO commits " \
              "(hash, datetime, message, author, additions, deletions)" \
              "VALUES (?,?,?,?,?,?)"

        commit_hash = commit.hexsha
        datetime = commit.committed_datetime.isoformat()
        message = message
        author = commit.author.name
        additions = commit.stats.total['insertions']
        deletions = commit.stats.total['deletions']

        values = (commit_hash, datetime, message, author, additions, deletions)

        cur.execute(sql, values)

        con.commit()
        con.close()

    def insert_into_tokens_files(self, commit, tokens):
        con = sqlite3.connect(self.db_location)
        cur = con.cursor()

        files_sql = "INSERT OR IGNORE INTO files " \
                    "(file_id, filepath)" \
                    "VALUES (?,?)"

        file_list = list(commit.stats.files.keys())
        files_values = [(uuid(), filepath) for filepath in file_list]
        cur.executemany(files_sql, files_values)

        file_ids = cur.execute(
            "SELECT file_id FROM files WHERE filepath in ({0})".format(', '.join('?' for _ in file_list)),
            file_list).fetchall()

        tokens_sql = "INSERT OR IGNORE INTO tokens " \
                     "(token_id, token_string)" \
                     "VALUES (?,?)"

        tokens_values = [(uuid(), token) for token in tokens]
        cur.executemany(tokens_sql, tokens_values)

        token_ids = cur.execute(
            "SELECT token_id FROM tokens WHERE token_string in ({0})".format(', '.join('?' for _ in tokens)),
            tokens).fetchall()

        tokens_files_values = [(uuid(), token_id[0], file_id[0]) for token_id in token_ids for file_id in file_ids]
        cur.executemany("INSERT INTO tokens_files (token_file_id, token_id, file_id) VALUES (?,?,?)",
                        tokens_files_values)

        con.commit()
        con.close()

    def meta_update(self, start_time, end_time, start_rev, end_rev, number_of_commits):
        con = sqlite3.connect(self.db_location)
        cur = con.cursor()

        sql = "INSERT INTO updates " \
              "(update_id, start_time, end_time, start_rev, end_rev, number_of_commits) " \
              "VALUES (?,?,?,?,?,?)"

        values = (uuid(), start_time.isoformat(), end_time.isoformat(), start_rev, end_rev, number_of_commits)

        cur.execute(sql, values)

        con.commit()
        con.close()

    def create_if_not_exists_tables(self):
        con = sqlite3.connect(self.db_location)
        cur = con.cursor()

        cur.execute("CREATE TABLE IF NOT EXISTS commits ("
                    "hash BLOB PRIMARY KEY, "
                    "datetime TEXT, "
                    "message TEXT, "
                    "author TEXT,"
                    "additions INTEGER,"
                    "deletions INTEGER"
                    ")")

        cur.execute("CREATE TABLE IF NOT EXISTS tokens ("
                    "token_id BLOB PRIMARY KEY,"
                    "token_string TEXT UNIQUE "
                    ")")

        cur.execute("CREATE TABLE IF NOT EXISTS files ("
                    "file_id BLOB PRIMARY KEY,"
                    "filepath TEXT UNIQUE"
                    ")")

        cur.execute("CREATE TABLE IF NOT EXISTS tokens_files ("
                    "token_file_id BLOB PRIMARY KEY,"
                    "token_id BLOB,"
                    "file_id BLOB,"
                    "tf_idf_score REAL,"
                    "FOREIGN KEY(token_id) REFERENCES tokens(token_id),"
                    "FOREIGN KEY (file_id) REFERENCES files(file_id)"
                    ")")

        cur.execute("CREATE TABLE IF NOT EXISTS updates ("
                    "update_id BLOB PRIMARY KEY,"
                    "start_time TEXT,"
                    "end_time TEXT,"
                    "start_rev TEXT,"
                    "end_rev TEXT,"
                    "number_of_commits INTEGER"
                    ")")

        con.commit()
        con.close()

    def get_last_indexed_rev(self):
        con = sqlite3.connect(self.db_location)
        cur = con.cursor()

        last_rev = cur.execute("SELECT end_rev FROM updates ORDER BY end_time LIMIT 1").fetchone()[0]

        con.commit()
        con.close()

        return last_rev

    def get_last_updated(self):
        con = sqlite3.connect(self.db_location)
        cur = con.cursor()

        last_updated = cur.execute("SELECT end_time FROM updates ORDER BY end_time LIMIT 1").fetchone()[0]

        con.commit()
        con.close()

        return last_updated

    def search(self, query):
        con = sqlite3.connect(self.db_location)
        cur = con.cursor()

        results = cur.execute(
            "SELECT tokens.token_string, f.filepath "
            "FROM tokens "
            "JOIN tokens_files tf "
            "ON tokens.token_id = tf.token_id "
            "JOIN files f "
            "ON tf.file_id = f.file_id "
            "WHERE token_string like :query "
            "GROUP BY f.filepath",
            {"query": query}).fetchall()

        con.commit()
        con.close()

        keys = ["keywords", "filepath"]
        return [dict(zip(keys, result)) for result in results]



from datetime import datetime
from git import Repo
from .database import Database
from .nlp_utils import preprocess_message


class RosettaIndex:
    def __init__(self, repo_path, db_path=None):
        self.repo_obj = Repo(repo_path)
        self.database = Database(db_path)
        if self.database.index_is_empty():
            raise Exception("Index is empty")

    def build_index(self):
        start_time = datetime.now()
        num_commits = 0
        for commit in self.repo_obj.iter_commits():
            start_rev = commit.hexsha
            num_commits += 1
            tokens, message = preprocess_message(commit.message)
            self.database.insert_into_commits(commit, message)
            self.database.insert_into_tokens_files(commit, tokens)
        end_rev = self.repo_obj.head.commit.hexsha
        end_time = datetime.now()
        self.database.meta_update(start_time, end_time, start_rev, end_rev, num_commits)
        print("Initial indexing complete. \n {} commits indexed in {} seconds".format(num_commits, (
                end_time - start_time).total_seconds()))

    def update_index(self):
        rev_list_arg = '{}..HEAD'.format(self.database.get_last_indexed_rev())
        start_time = datetime.now()
        num_commits = 0
        start_rev = None
        for commit in self.repo_obj.iter_commits(rev_list_arg):
            start_rev = commit.hexsha
            num_commits += 1
            tokens, message = preprocess_message(commit.message)
            self.database.insert_into_commits(commit, message)
            self.database.insert_into_tokens_files(commit, tokens)
        end_rev = self.repo_obj.head.commit.hexsha
        end_time = datetime.now()
        if start_rev is not None:
            self.database.meta_update(start_time, end_time, start_rev, end_rev, num_commits)
        return num_commits, end_time

    def search_index(self, query):
        return self.database.search(query)

    def get_last_updated(self):
        return self.database.get_last_updated()



import json
import os
import fnmatch
import sqlite3
from urllib.request import pathname2url
import click
from git import Repo


def json_echo(dictionary):
    click.echo(json.dumps(dictionary, indent=2))


def normalise_paths(*paths):
    results = []
    for path in paths:
        if path is not None:
            results.append(os.path.normpath(path))
        else:
            results.append(path)
    return tuple(results)


def check_index_path(repo_path, index_path=None):
    if not index_path:
        index_path = look_for_existing_index(repo_path)
    if not index_path:
        raise click.UsageError("No existing index found. Use create to create an index.")
    try:
        check_db_path(index_path)
    except Exception:
        raise click.UsageError("Index path provided is not a valid Rosetta Index.")
    return index_path


def check_repo_path(repo_path):
    if repo_path is None:
        raise click.UsageError("No repo path provided")
    repo_obj = Repo(repo_path)
    if repo_obj.bare:
        raise click.UsageError(f"No git repo found at {repo_path}")


def look_for_existing_index(path):
    for root, dirs, files in os.walk(path):
        for filename in files:
            if fnmatch.fnmatch(filename, "*.db"):
                return os.path.join(root, filename)


def check_db_path(db_path):
    try:
        database_uri = f"file:{pathname2url(db_path)}?mode=rw"
        conn = sqlite3.connect(database_uri, uri=True)
    except sqlite3.OperationalError:
        raise Exception(f"{db_path} is not a valid Rosetta index.")



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
setup(
    name='rosetta_search',
    url="https://github.com/mnckapilan/rosetta_search",
    author="Kapilan M",
    author_email="mnckapilan+git@gmail.com",
    version='0.0.6',
    description="Semantic Code Search informed by git commit messages",
    packages=find_packages(where='src'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    package_dir={'': 'src'},
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "GitPython~=3.1.14",
        "nltk~=3.6.1",
        "click~=7.1.2"
    ],
    entry_points='''
    [console_scripts]
    rosetta=rosetta_search:rosetta
    '''
)
